package com.bocloud.gateway.exception;

import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.autoconfigure.web.ErrorProperties;
import org.springframework.boot.autoconfigure.web.WebProperties;
import org.springframework.boot.autoconfigure.web.reactive.error.DefaultErrorWebExceptionHandler;
import org.springframework.boot.web.error.ErrorAttributeOptions;
import org.springframework.boot.web.reactive.error.ErrorAttributes;
import org.springframework.cloud.client.discovery.DiscoveryClient;
import org.springframework.context.ApplicationContext;
import org.springframework.web.reactive.function.server.*;
import org.springframework.web.server.ResponseStatusException;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

@Slf4j
public class GlobalExceptionHandler extends DefaultErrorWebExceptionHandler {
    private final DiscoveryClient discoveryClient;

    public GlobalExceptionHandler(ErrorAttributes errorAttributes, WebProperties.Resources resources,
                                  ErrorProperties errorProperties, DiscoveryClient discoveryClient,
                                  ApplicationContext applicationContext) {
        super(errorAttributes, resources, errorProperties, applicationContext);
        this.discoveryClient = discoveryClient;
    }

    /**
     * 构建返回的JSON数据格式
     *
     * @param status 状态码
     * @param error  异常信息
     * @return 返回数据
     */
    public static Map<String, Object> response(int status, String error) {
        Map<String, Object> response = new HashMap<>();
        response.put("code", status);
        response.put("message", error);
        response.put("success", false);
        response.put("failed", true);
        return response;
    }

    /**
     * 获取异常属性
     */
    @Override
    protected Map<String, Object> getErrorAttributes(ServerRequest request, ErrorAttributeOptions options) {
        int code = 500;
        String message;
        Throwable error = super.getError(request);
        if (error instanceof ResponseStatusException) {
            code = ((ResponseStatusException) error).getStatusCode().value();
        }
        switch (code) {
            case 404:
                message = ":服务路由不存在";
                break;
            case 503:
                message = ":服务实例不存在";
                break;
            default:
                message = ":服务实例维护中";
                break;
        }
        Map<String, Object> response = response(code, "网关异常" + message);
        response.put("data", this.buildMessage(request, error));
        return response;
    }

    /**
     * 指定响应处理方法为JSON处理的方法
     *
     * @param errorAttributes 异常属性
     */
    @Override
    protected RouterFunction<ServerResponse> getRoutingFunction(ErrorAttributes errorAttributes) {
        return RouterFunctions.route(RequestPredicates.all(), this::renderErrorResponse);
    }

    /**
     * 根据code获取对应的HttpStatus
     *
     * @param errorAttributes 异常属性
     */
    @Override
    protected int getHttpStatus(Map<String, Object> errorAttributes) {
        return (int) errorAttributes.get("code");
    }

    /**
     * 构建异常信息
     *
     * @param request   请求
     * @param exception 异常
     * @return 消息
     */
    private String buildMessage(ServerRequest request, Throwable exception) {
        StringBuilder message = new StringBuilder("Failed to handle ");
        String path = request.exchange().getRequest().getPath().value();
        List<String> services = this.discoveryClient.getServices();
        if (path.startsWith("/api/")) {
            String service = path.split("/")[2];
            if (services.contains(service)) {
                message.append("service [").append(service).append("] request [");
            } else {
                message.append("service [").append(service).append("] request [");
            }
        } else {
            message.append("request [");
        }
        message.append(request.methodName());
        message.append(" ");
        message.append(request.uri());
        message.append("] for reason");
        if (Objects.nonNull(exception)) {
            message.append(": ");
            message.append(exception.getMessage());
        }
        log.error("gateway error : {}", message);
        return message.toString();
    }

}
