package com.bocloud.gateway.exception;

import com.alibaba.fastjson.JSONObject;
import com.google.common.base.Joiner;
import com.google.common.base.Throwables;
import com.google.common.collect.Lists;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.reactivestreams.Publisher;
import org.springframework.cloud.client.ServiceInstance;
import org.springframework.cloud.client.loadbalancer.Response;
import org.springframework.cloud.gateway.filter.GatewayFilterChain;
import org.springframework.cloud.gateway.filter.GlobalFilter;
import org.springframework.core.Ordered;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DataBufferFactory;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.http.HttpStatus;
import org.springframework.http.HttpStatusCode;
import org.springframework.http.server.reactive.ServerHttpRequest;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.http.server.reactive.ServerHttpResponseDecorator;
import org.springframework.stereotype.Component;
import org.springframework.web.server.ServerWebExchange;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Objects;

import static org.springframework.cloud.gateway.support.ServerWebExchangeUtils.GATEWAY_LOADBALANCER_RESPONSE_ATTR;
import static org.springframework.cloud.gateway.support.ServerWebExchangeUtils.ORIGINAL_RESPONSE_CONTENT_TYPE_ATTR;

@Slf4j
@Component
public class GlobalResponseFilter implements GlobalFilter, Ordered {
    private static final Joiner joiner = Joiner.on("");

    @Override
    public Mono<Void> filter(ServerWebExchange exchange, GatewayFilterChain chain) {
        ServerHttpRequest originalRequest = exchange.getRequest();
        ServerHttpResponse originalResponse = exchange.getResponse();
        DataBufferFactory bufferFactory = originalResponse.bufferFactory();
        ServerHttpResponseDecorator response = new ServerHttpResponseDecorator(originalResponse) {
            @Override
            @NonNull
            public Mono<Void> writeWith(@NonNull Publisher<? extends DataBuffer> body) {
                if (Objects.equals(getStatusCode(), HttpStatus.OK) && body instanceof Flux) {
                    return super.writeWith(body);
                } else {
                    // 获取ContentType，判断是否返回JSON格式数据
                    String originalResponseContentType = exchange.getAttribute(ORIGINAL_RESPONSE_CONTENT_TYPE_ATTR);
                    if (StringUtils.isNotBlank(originalResponseContentType) && originalResponseContentType.contains("application/json")) {
                        Flux<? extends DataBuffer> fluxBody = Flux.from(body);
                        //（返回数据内如果字符串过大，默认会切割）解决返回体分段传输
                        return super.writeWith(fluxBody.buffer().map(dataBuffers -> {
                            List<String> list = Lists.newArrayList();
                            dataBuffers.forEach(dataBuffer -> {
                                try {
                                    byte[] content = new byte[dataBuffer.readableByteCount()];
                                    dataBuffer.read(content);
                                    DataBufferUtils.release(dataBuffer);
                                    list.add(new String(content, StandardCharsets.UTF_8));
                                } catch (Exception e) {
                                    log.error("加载Response字节流异常，失败原因：{}", Throwables.getStackTraceAsString(e));
                                }
                            });
                            String responseData = joiner.join(list);
                            Response<ServiceInstance> loadBalanceResponse = exchange.getAttribute(GATEWAY_LOADBALANCER_RESPONSE_ATTR);
                            responseData = responseHandle(responseData, getStatusCode(), pluginJudge(originalRequest, loadBalanceResponse));
                            byte[] uppedContent = new String(responseData.getBytes(), StandardCharsets.UTF_8).getBytes();
                            originalResponse.getHeaders().setContentLength(uppedContent.length);
                            return bufferFactory.wrap(uppedContent);
                        }));
                    } else {
                        return super.writeWith(body);
                    }
                }
            }

            private RequestPathEntry pluginJudge(ServerHttpRequest request, Response<ServiceInstance> response) {
                String name;
                String path = request.getPath().value();
                boolean isPlugin = path.contains("/cmp/plugins/") || path.contains("/cop/plugins/") || path.contains("/cos/plugins/");
                if (isPlugin) {
                    name = path.split("/")[4];
                } else {
                    name = path.split("/")[2];
                }
                String host = response.hasServer() ? response.getServer().getHost() : "unknown";
                Integer port = response.hasServer() ? response.getServer().getPort() : 0;
                return RequestPathEntry.builder()
                        .plugin(isPlugin)
                        .name(name)
                        .host(host)
                        .port(port)
                        .path(request.getPath().value())
                        .build();
            }

            /**
             * 返回数据处理
             *
             * @param responseData 响应数据
             * @param httpStatusCode 响应状态
             * @param entry 请求信息
             * @return 处理数据结果
             */
            private String responseHandle(String responseData, HttpStatusCode httpStatusCode, RequestPathEntry entry) {
                String responseJson;
                try {
                    JSONObject jsonObject = JSONObject.parseObject(responseData);
                    jsonObject.put("message", buildMessage(httpStatusCode, entry));
                    responseJson = jsonObject.toJSONString();
                    log.warn("error: {}", jsonObject.getString("message"));
                } catch (Exception e) {
                    log.error("返回数据处理转化失败，异常信息={}", e.getMessage());
                    return responseData;
                }
                return responseJson;
            }

            private String buildMessage(HttpStatusCode httpStatusCode, RequestPathEntry entry) {
                String message;
                HttpStatus status = HttpStatus.valueOf(httpStatusCode.value());
                switch (status) {
                    case NOT_FOUND:
                        if (entry.isPlugin()) {
                            message = entry.getName().toUpperCase() + "插件实例" + entry.getHost() + ":" + entry.getPort() + "未安装或未启动";
                        } else {
                            message = "请求路径在" + entry.getName().toUpperCase() + "服务实例" + entry.getHost() + ":" + entry.getPort() + "上不存在";
                        }
                        break;
                    case BAD_REQUEST:
                        message = "参数异常";
                        break;
                    case UNAUTHORIZED:
                        message = "请求禁止";
                        break;
                    case GATEWAY_TIMEOUT:
                        message = "网关超时";
                        break;
                    case METHOD_NOT_ALLOWED:
                        message = "请求方法不允许";
                        break;
                    case INTERNAL_SERVER_ERROR:
                        if (entry.isPlugin()) {
                            message = entry.getName().toUpperCase() + "插件实例" + entry.getHost() + ":" + entry.getPort() + "内部异常";
                        } else {
                            message = entry.getName().toUpperCase() + "服务实例" + entry.getHost() + ":" + entry.getPort() + "内部异常";
                        }
                        break;
                    case LOOP_DETECTED:
                        message = "循环调用";
                        break;
                    default:
                        message = "未知异常";
                }
                return message;
            }

            @Override
            @NonNull
            public Mono<Void> writeAndFlushWith(@NonNull Publisher<? extends Publisher<? extends DataBuffer>> body) {
                return writeWith(Flux.from(body).flatMapSequential(p -> p));
            }
        };
        return chain.filter(exchange.mutate().response(response).build());
    }

    @Override
    public int getOrder() {
        // -1 is response write filter, must be called before that
        return -2;
    }
}
