package com.bocloud.gateway.route;

import com.megatron.framework.core.CurrentService;
import com.megatron.framework.registry.RegistryProperties;
import jakarta.annotation.PreDestroy;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.curator.RetryPolicy;
import org.apache.curator.framework.CuratorFramework;
import org.apache.curator.framework.CuratorFrameworkFactory;
import org.apache.curator.framework.api.ACLProvider;
import org.apache.curator.framework.recipes.cache.CuratorCache;
import org.apache.curator.framework.recipes.cache.CuratorCacheListener;
import org.apache.curator.framework.recipes.cache.PathChildrenCacheEvent;
import org.springframework.beans.factory.InitializingBean;
import org.springframework.cloud.zookeeper.ZookeeperProperties;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;


/**
 * @author dmw
 */
@Slf4j
@Component
@RequiredArgsConstructor
public class GatewayRouteDaemon implements InitializingBean {

    private final CurrentService currentService;
    private final GatewayRouteHandler routeHandler;
    private final ACLProvider aclProvider;
    private final RetryPolicy retryPolicy;
    private final ZookeeperProperties properties;
    private final RegistryProperties registryProperties;

    private CuratorCache serviceWatcher;

    @Override
    public void afterPropertiesSet() throws Exception {
        String watchPath = this.currentService.getService().getPrefix();
        log.info("watch path is {}", watchPath);
        CuratorFramework curatorFramework = buildClient();
        this.serviceWatcher = CuratorCache.build(curatorFramework, watchPath);
        CuratorCacheListener listener = CuratorCacheListener.builder()
                .forPathChildrenCache(watchPath, this.currentService.getClient().getClient(), (client, event) -> refreshRoutes(event))
                .build();
        this.serviceWatcher.listenable().addListener(listener);
        this.serviceWatcher.start();
    }

    @PreDestroy
    public void destroy() {
        this.serviceWatcher.close();
    }

    private CuratorFramework buildClient() throws InterruptedException {
        CuratorFrameworkFactory.Builder builder = CuratorFrameworkFactory.builder();
        if (registryProperties.isSecurity() && StringUtils.hasText(registryProperties.getUsername())
                && StringUtils.hasText(registryProperties.getPassword())) {
            builder.authorization("digest", registryProperties.getUserPassword().getBytes());
        }
        builder.connectString(properties.getConnectString());
        builder.connectionTimeoutMs(registryProperties.getConnectTimeout());
        builder.sessionTimeoutMs(registryProperties.getSessionTimeout());
        builder.aclProvider(aclProvider);
        CuratorFramework curator = builder.retryPolicy(retryPolicy).build();
        curator.start();
        curator.blockUntilConnected(properties.getBlockUntilConnectedWait(), properties.getBlockUntilConnectedUnit());
        return curator;
    }

    private void refreshRoutes(PathChildrenCacheEvent event) {
        if (event.getType().equals(PathChildrenCacheEvent.Type.CHILD_ADDED) || event.getType().equals(PathChildrenCacheEvent.Type.CHILD_REMOVED)) {
            String path = event.getData().getPath();
            String[] levels = path.split("/");
            if (levels.length == 6) {
                log.info("refresh routes for path {} event: {}", event.getData().getPath(), event.getType());
                this.routeHandler.load();
            }
        }
    }
}
