package com.bocloud.gateway.route;

import com.alibaba.fastjson.JSONArray;
import com.bocloud.gateway.domain.FiltersEntity;
import com.bocloud.gateway.domain.GatewayRequest;
import com.bocloud.gateway.domain.GatewayRoute;
import com.bocloud.gateway.domain.PredicatesEntity;
import com.megatron.framework.core.CurrentService;
import lombok.Getter;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.CommandLineRunner;
import org.springframework.cloud.client.discovery.DiscoveryClient;
import org.springframework.cloud.gateway.event.RefreshRoutesEvent;
import org.springframework.cloud.gateway.filter.FilterDefinition;
import org.springframework.cloud.gateway.handler.predicate.PredicateDefinition;
import org.springframework.cloud.gateway.route.RouteDefinition;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.ApplicationEventPublisherAware;
import org.springframework.stereotype.Service;
import org.springframework.web.util.UriComponentsBuilder;
import reactor.core.publisher.Mono;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * 动态路由服务，用来实现路由的添加，更新，删除和查询操作
 *
 * @author dmw
 */
@Slf4j
@Service
public class GatewayRouteHandler implements ApplicationEventPublisherAware, CommandLineRunner {

    private final CurrentService currentService;
    private final DiscoveryClient discoveryClient;
    @Getter
    private final LocalRouteRepository localRouteRepository;
    private ApplicationEventPublisher publisher;

    @Autowired
    public GatewayRouteHandler(CurrentService currentService, DiscoveryClient discoveryClient, LocalRouteRepository localRouteRepository) {
        this.currentService = currentService;
        this.discoveryClient = discoveryClient;
        this.localRouteRepository = localRouteRepository;
    }

    @Override
    public void setApplicationEventPublisher(@NonNull ApplicationEventPublisher applicationEventPublisher) {
        this.publisher = applicationEventPublisher;
    }

    /**
     * 增加一条路由
     *
     * @param gatewayRoute 网关路由信息
     */
    public void addRoute(GatewayRoute gatewayRoute) {
        RouteDefinition definition = handleData(gatewayRoute);
        this.localRouteRepository.save(Mono.just(definition)).subscribe();
        this.publisher.publishEvent(new RefreshRoutesEvent(this));
        log.info("add route {} success", gatewayRoute.getServiceName());
    }

    /**
     * 更新一条路由信息
     *
     * @param gatewayRoute 路由信息对象
     */
    public void updateRoute(GatewayRoute gatewayRoute) {
        RouteDefinition definition = handleData(gatewayRoute);
        try {
            this.localRouteRepository.update(Mono.just(definition)).subscribe();
            this.publisher.publishEvent(new RefreshRoutesEvent(this));
            log.info("modify route {} success", gatewayRoute.getServiceName());
        } catch (Exception e) {
            log.error("modify route exception!", e);
        }
    }

    /**
     * 删除一条路由信息
     *
     * @param routeId 路由ID
     */
    public void deleteRoute(String routeId) {
        this.localRouteRepository.delete(Mono.just(routeId)).subscribe();
        this.publisher.publishEvent(new RefreshRoutesEvent(this));
        log.info("delete route {} success", routeId);
    }

    /**
     * Callback used to run the bean.
     *
     * @param args incoming main method arguments
     * @throws Exception on error
     */
    @Override
    public void run(String... args) throws Exception {
        this.load();
    }

    /**
     * 加载路由配置
     */
    public void load() {
        log.info("start to load route from zookeeper ...");
        List<String> services = this.discoveryClient.getServices();
        services.stream().filter(service -> !service.equalsIgnoreCase(currentService.getService().getName()))
                .forEach(this::handleRoute);
        this.publisher.publishEvent(new RefreshRoutesEvent(this));
        log.info("load route from zookeeper success !!!");
    }

    private void handleRoute(String service) {
        GatewayRoute gatewayRoute = new GatewayRequest(service).toGatewayRoute();
        Mono<RouteDefinition> mono = Mono.just(this.handleData(gatewayRoute));
        this.localRouteRepository.save(mono).subscribe();
    }

    /**
     * 路由数据转换公共方法
     *
     * @param gatewayRoute 网关路由对象
     * @return 路由定义
     */
    private RouteDefinition handleData(GatewayRoute gatewayRoute) {
        RouteDefinition definition = new RouteDefinition();
        URI uri;
        String http = "http";
        if (gatewayRoute.getUri().startsWith(http)) {
            //http地址
            uri = UriComponentsBuilder.fromHttpUrl(gatewayRoute.getUri()).build().toUri();
        } else {
            //注册中心
            String lb = "lb://";
            uri = UriComponentsBuilder.fromUriString(lb + gatewayRoute.getUri()).build().toUri();
        }
        definition.setId(gatewayRoute.getServiceId());
        // 获取过滤器json字符串
        String filters = gatewayRoute.getFilters();
        //获取断言json字符串
        String predicates = gatewayRoute.getPredicates();
        List<FiltersEntity> filtersEntities;
        List<PredicatesEntity> predicatesEntities;
        filtersEntities = JSONArray.parseArray(filters, FiltersEntity.class);
        predicatesEntities = JSONArray.parseArray(predicates, PredicatesEntity.class);
        //循环拼装断言信息
        List<PredicateDefinition> predicateDefinitions = new ArrayList<>();
        String dot = ",";
        String path = "Path";
        String key = "_genkey_";
        String key0 = "_genkey_0";
        predicatesEntities.forEach(entity -> {
            PredicateDefinition predicate = new PredicateDefinition();
            Map<String, String> predicateParams = new HashMap<>(predicatesEntities.size());
            // 名称是固定的，spring gateway会根据名称找对应的PredicateFactory
            predicate.setName(entity.getType());
            //判断断言有分号则进行拼接
            if (path.equalsIgnoreCase(entity.getType())) {
                String pattern = "pattern";
                if (entity.getPredicates().contains(dot)) {
                    String[] predicateArray = entity.getPredicates().split(dot);
                    for (int i = 0; i < predicateArray.length; i++) {
                        predicateParams.put(pattern + (i + 1), predicateArray[i]);
                    }
                } else {
                    predicateParams.put(pattern, entity.getPredicates());
                }
            } else {
                if (entity.getPredicates().contains(dot)) {
                    String[] predicateArray = entity.getPredicates().split(dot);
                    for (int i = 0; i < predicateArray.length; i++) {
                        predicateParams.put(key + i, predicateArray[i]);
                    }
                } else {
                    predicateParams.put(key0, entity.getPredicates());
                }
            }
            predicate.setArgs(predicateParams);
            predicateDefinitions.add(predicate);
        });
        //循环拼装过滤器信息
        List<FilterDefinition> filterDefinitions = new ArrayList<>();
        filtersEntities.forEach(filter -> {
            // 名称是固定的, 路径去前缀
            FilterDefinition filterDefinition = new FilterDefinition();
            Map<String, String> filterParams = new HashMap<>(filtersEntities.size());
            //设置过滤器名称
            filterDefinition.setName(filter.getType());
            //判断过滤器如果有逗号则进行分割
            if (filter.getRule().contains(dot)) {
                String[] rules = filter.getRule().split(dot);
                for (int i = 0; i < rules.length; i++) {
                    filterParams.put(key + i, rules[i]);
                }
            } else {
                filterParams.put(key0, filter.getRule());
            }
            filterDefinition.setArgs(filterParams);
            filterDefinitions.add(filterDefinition);
        });
        definition.setPredicates(predicateDefinitions);
        definition.setFilters(filterDefinitions);
        definition.setUri(uri);
        definition.setOrder(gatewayRoute.getOrder());
        return definition;
    }
}
